r"""

DNS
###

**Direct Numerical Simulation**

This module contains routines for performing direct numerical
simulations.

.. currentmodule:: porespy

.. autosummary::
   :template: mybase.rst
   :toctree: generated/

   dns.tortuosity

"""

__all__ = ["tortuosity"]

from ._funcs import tortuosity
